package edu.sc.seis.cormorant.seismogram;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.TimeZone;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.RollingFileAppender;

import sun.net.ftp.FtpProtocolException;
import edu.iris.Fissures.model.ISOTime;
import edu.iris.Fissures.model.MicroSecondDate;
import edu.sc.seis.fissuresUtil.cache.WorkerThreadPool;
import edu.sc.seis.fissuresUtil.chooser.ClockUtil;
import edu.sc.seis.seisFile.mseed.DataHeader;
import edu.sc.seis.seisFile.mseed.DataRecord;
import edu.sc.seis.seisFile.mseed.SeedFormatException;
import edu.sc.seis.seisFile.mseed.SeedRecord;

/**
 * @author crotwell
 * 
 *         Created on Nov 30, 2010
 */
public class BudLightPond implements Runnable {

    public BudLightPond(DataRecordQueue queue, File dataDir) {
        this.queue = queue;
        this.dataDir = dataDir;
        pondSyncDir = new File(dataDir, POND_SYNC_DATA);
        if (!pondSyncDir.exists()) {
            pondSyncDir.mkdirs();
        }
        dataPool.setMaxQueueSize(20);
        netPool.setMaxQueueSize(10);
        eventPool.setMaxQueueSize(10);
        dirPool.setMaxQueueSize(10);
    }

    public void run() {
        process();
    }

    public void process() {
        try {
            if (eventDirUrl != null) {
                logger.info("Process: "+eventDirUrl);
                processEvent(eventDirUrl);
            } else {
                logger.info("Process by month starting: "+beginYear+"-"+beginMonth);
            GregorianCalendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
            cal.setTime(ClockUtil.now());
            int curYear = cal.get(GregorianCalendar.YEAR);
            int curMonth = cal.get(GregorianCalendar.MONTH)+1;
            // so we can rest to Jan for following year
            int tmpBeginMonth = beginMonth; 
            int tmpEndMonth = 12;
            for (int year = beginYear; year <= endYear && year <= curYear; year++) {
                if (year == endYear) {tmpEndMonth = endMonth;}
                for (int month = tmpBeginMonth; month <= 12 && month <= tmpEndMonth && (year < curYear || month <= curMonth); month++) {
                    processMonth(year, month);
                }
                if (year == beginYear) {
                    tmpBeginMonth = 1; // reset to Jan for year after beginYear
                }
            }
            }
            while (dataPool.isEmployed()) {
                sleep(1);
            }
            System.out.println("Done");
        } catch(MalformedURLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch(IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    void processMonth(int year, int month) throws IOException {
        final String monthdir = POND_BASE_URL + "/" + year + "/" + BudLightWriter.twoDigit.format(month) + "/";
        List<IrisDirListing> files = dirListing(monthdir);
        for (final IrisDirListing irisDirListing : files) {
            System.out.println("Event: " + irisDirListing.filename);
            if (irisDirListing.perm.startsWith("dr") && irisDirListing.filename.endsWith(".farm")) {
                eventPool.invokeLater(new Runnable() {

                    public void run() {
                        String url = monthdir + irisDirListing.filename + "/";
                        try {
                            processEvent(url);
                        } catch(Throwable e) {
                            logger.error(url, e);
                        }
                    }
                });
            }
        }
    }

    void processEvent(final String eventDir) throws IOException {
        List<IrisDirListing> files = dirListing(eventDir);
        for (final IrisDirListing irisDirListing : files) {
            System.out.println("   Net: " + irisDirListing.filename);
            if (irisDirListing.perm.startsWith("dr")) {
                netPool.invokeLater(new Runnable() {

                    public void run() {
                        String url = eventDir + irisDirListing.filename + "/";
                        try {
                            processNetwork(url);
                        } catch(Throwable e) {
                            logger.error(url, e);
                        }
                    }
                });
            }
        }
    }

    void processNetwork(final String dir) throws IOException {
        List<IrisDirListing> files = dirListing(dir);
        for (final IrisDirListing irisDirListing : files) {
            if (irisDirListing.perm.startsWith("-r") && irisDirListing.filename.startsWith("miniSEED.")) {
                File pondInfoDir = new File(pondSyncDir, dir.substring(POND_BASE_URL.length() + 1));
                final File pondInfo = new File(pondInfoDir, RETRIEVAL_TIME_FILE);
                if (!pondInfoDir.exists()) {
                    pondInfoDir.mkdirs();
                } else if (pondInfo.exists()) {
                    BufferedReader syncRead = new BufferedReader(new FileReader(pondInfo));
                    String line = null;
                    String s;
                    // get last line in file for most recent data access
                    while ((s = syncRead.readLine()) != null) {
                        line = s;
                    }
                    if (line != null) {
                        int firstComma = line.indexOf(",");
                        MicroSecondDate lastRetrieval = new MicroSecondDate(line.substring(0, firstComma));
                        System.out.println("        Last retrieval: " + lastRetrieval);
                        if (lastRetrieval.after(irisDirListing.getModTime())) {
                            // do not need to reget this one
                            continue;
                        }
                    }
                }
                totalSize += irisDirListing.size;
                System.out.println("      miniseed: " + irisDirListing.filename + "  " + irisDirListing.size + " "
                        + totalSize / 1024 / 1024 + " Mb");
                while (queue.getSize() > 10000) {sleep(1);}
                dataPool.invokeLater(new Runnable() {

                    public void run() {
                        try {
                            retrieveData(dir + irisDirListing.filename, 3);
                            // must be ok
                            BufferedWriter syncData = new BufferedWriter(new FileWriter(pondInfo, true));
                            syncData.write(ISOTime.getISOString(ClockUtil.now()) + "," + irisDirListing.toCSV());
                            syncData.newLine();
                            syncData.close();
                        } catch(Throwable e) {
                            logger.error(dir + irisDirListing.filename, e);
                            e.printStackTrace();
                        }
                    }
                });
            }
        }
    }

    int totalSize = 0;

    public List<IrisDirListing> dirListing(String urlString) throws IOException {
        return dirListing(urlString, 3);
    }

    public List<IrisDirListing> dirListing(String urlString, int retries) throws IOException {
        BufferedReader in = null;
        try {
            URL url = new URL(urlString);
            URLConnection conn = url.openConnection();
            in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
            String line;
            List<IrisDirListing> out = new ArrayList<IrisDirListing>();
            while ((line = in.readLine()) != null) {
                // System.out.println("Line: " + line);
                String[] items = line.split(" +");
                if (items.length > 8) {
                    // System.out.println("Filename: " + items[8]);
                    out.add(new IrisDirListing(items));
                }
            }
            in.close();
            return out;
        } catch(FtpProtocolException e) {
            if (in != null) {
                in.close();
                in = null;
            }
            if (retries <= 0) {
                logger.error("no retry " + urlString, e);
                writeError("FtpProtocolException", e, urlString);
                throw e;
            } else {
                logger.warn("retrying " + retries + " " + urlString, e);
                sleep(1);
                return dirListing(urlString, retries - 1);
            }
        }
    }

    boolean retrieveData(String urlString) {
        return retrieveData(urlString, 3);
    }

    boolean retrieveData(String urlString, int retries) {
        DataInputStream dataInStream = null;
        boolean firstRecord = true;
        SeedRecord sr;
        try {
            URL url = new URL(urlString);
            URLConnection conn = url.openConnection();
            dataInStream = new DataInputStream(new BufferedInputStream(conn.getInputStream()));
            while ((sr = SeedRecord.read(dataInStream, 4096)) != null) {
                if (sr instanceof DataRecord) {
                    DataRecord dr = (DataRecord)sr;
                    if (firstRecord) {
                        if (dr.getBlockettes(1000).length == 0) {
                            // no b1000
                            File pondInfoDir = new File(pondSyncDir, urlString.substring(POND_BASE_URL.length() + 1)).getParentFile();
                            File pondInfo = getMissingB1000File(pondInfoDir, dr);
                            PrintWriter pondInfoOut = new PrintWriter(new BufferedWriter(new FileWriter(pondInfo, true)));
                            dr.writeASCII(pondInfoOut, "  ");
                            pondInfoOut.close();
                            System.out.println("missing B1000 pondInfo: " + pondInfo);
                        }
                    }
                    queue.addWork(dr);
                }
            }
            return true;
        } catch(EOFException e) {
            // done
        } catch(FtpProtocolException e) {
            if (dataInStream != null) {
                try {
                    dataInStream.close();
                } catch(IOException e1) {}
                dataInStream = null;
            }
            if (retries <= 0) {
                logger.error("no retry " + urlString, e);
                writeError("FtpProtocolException", e, urlString);
            } else {
                logger.error("retrying " + retries + " " + urlString, e);
                sleep(1);
                return retrieveData(urlString, retries - 1);
            }
        } catch(IOException e) {
            logger.error("",e);
            writeError("IOException", e, urlString);
        } catch(SeedFormatException e) {
            logger.error("",e);
            writeError("SeedFormatException", e, urlString);
        } finally {
            if (dataInStream != null) {
                try {
                    dataInStream.close();
                } catch(IOException e) {
                    // oh well
                }
            }
        }
        return false;
    }

    void writeError(String message, Throwable t, String urlString) {
        PrintWriter out;
        try {
            File pondInfoDir = new File(pondSyncDir, urlString.substring(POND_BASE_URL.length() + 1));
            File pondInfo = new File(pondInfoDir, RETRIEVAL_ERROR_FILE);
            if (!pondInfoDir.exists()) {
                pondInfoDir.mkdirs();
            }
            out = new PrintWriter(new BufferedWriter(new FileWriter(pondInfo, true)));
            out.println(ISOTime.getISOString(ClockUtil.now()) + "" + message);
            out.println(t.getMessage());
            t.printStackTrace(out);
            out.close();
        } catch(IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    private static void sleep(int seconds) {
        try {
            Thread.sleep(seconds * 1000);
        } catch(InterruptedException e) {}
    }

    public static void main(String[] args) throws IOException {
    	RollingFileAppender r = new RollingFileAppender(new PatternLayout("%d{ISO8601} %p %t %c{2} - %m%n"),
                                                        "BudLightPond.log");
        r.setMaxBackupIndex(3);
        BasicConfigurator.configure(r);
        logger.info("Logging configured");
        String dataDir = "PondData";
        int bYear = -1;
        int bMonth = -1;
        int eYear = -1;
        int eMonth = -1;
        String url = null;
        for (int i = 0; i < args.length; i++) {
            if (i < args.length - 1) {
                if (args[i].equals("-d")) {
                    dataDir = args[i + 1];
                    i++;
                }
                if (args[i].equals("-y")) {
                    bYear = Integer.parseInt(args[i + 1]);
                    i++;
                }
                if (args[i].equals("-m")) {
                    bMonth = Integer.parseInt(args[i + 1]);
                    i++;
                }
                if (args[i].equals("-Y")) {
                    eYear = Integer.parseInt(args[i + 1]);
                    i++;
                }
                if (args[i].equals("-M")) {
                    eMonth = Integer.parseInt(args[i + 1]);
                    i++;
                }
                if (args[i].equals("--url")) {
                    url = args[i + 1];
                    i++;
                }
            }
        }
        DataRecordQueue queue = new DataRecordQueue();
        File dataFile = new File(dataDir);
        if (!dataFile.exists()) {
            dataFile.mkdirs();
        }
        BudLightWriter bdw = new BudLightWriter(dataFile, queue);
        Thread t = new Thread(bdw);
        t.setDaemon(true);
        t.start();
        BudLightPond blp = new BudLightPond(queue, dataFile);
        if (bYear != -1) { blp.setBeginYear(bYear);}
        if (bMonth != -1) {blp.setBeginMonth(bMonth);}
        if (eYear != -1) {blp.setEndYear(eYear);}
        if (eMonth != -1) {blp.setEndMonth(eMonth);}
        if (url != null) {blp.setEventDirUrl(url);}
        blp.run();
    }

    public int getBeginYear() {
        return beginYear;
    }

    public void setBeginYear(int beginYear) {
        this.beginYear = beginYear;
    }

    public int getEndYear() {
        return endYear;
    }

    public void setEndYear(int endYear) {
        this.endYear = endYear;
    }

    public int getBeginMonth() {
        return beginMonth;
    }

    public void setBeginMonth(int beginMonth) {
        this.beginMonth = beginMonth;
    }

    public int getEndMonth() {
        return endMonth;
    }

    public void setEndMonth(int endMonth) {
        this.endMonth = endMonth;
    }

    public void setEventDirUrl(String url) {
        this.eventDirUrl = url;
    }

    public static String filenameFromDirLine(String line) {
        System.out.println("Line is: " + line);
        return line.split("[\\\"]")[1];
    }

    public static File getMissingB1000File(File pondInfoDir, DataRecord dr) {
        DataHeader head = dr.getHeader();
        return new File(pondInfoDir, MISSING_B1000_FILE + head.getNetworkCode().trim() + "."
                + head.getStationIdentifier().trim() + "." + head.getLocationIdentifier().trim() + "."
                + head.getChannelIdentifier().trim());
    }

    public static final String RETRIEVAL_TIME_FILE = "retrievalTime.csv";

    public static final String RETRIEVAL_ERROR_FILE = "retrievalError";

    public static final String MISSING_B1000_FILE = "missingB1000_";

    public static final String POND_SYNC_DATA = ".pondSync";

    public static final String FTP_IRIS = "ftp.iris.washington.edu";

    public static final String POND_BASE_DIR = "pub/farm/POND";

    public static final String POND_BASE_URL = "ftp://ftp.iris.washington.edu/pub/farm/POND";

    WorkerThreadPool dirPool = new WorkerThreadPool("dir PondWorker", 2);

    WorkerThreadPool eventPool = new WorkerThreadPool("event PondWorker", 2);

    WorkerThreadPool netPool = new WorkerThreadPool("net PondWorker", 5);

    WorkerThreadPool dataPool = new WorkerThreadPool("data PondWorker", 3);

    File pondSyncDir;

    DataRecordQueue queue;

    File dataDir;

    int beginYear = 1990;

    int endYear = -1;

    int beginMonth = 1;

    int endMonth = 12;
    
    String eventDirUrl = null;

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(BudLightPond.class);
}
