package edu.sc.seis.cormorant.seismogram;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.seismogramDC.LocalSeismogramImpl;
import edu.sc.seis.fissuresUtil.comparator.DataRecordBeginComparator;
import edu.sc.seis.fissuresUtil.mseed.FissuresConvert;
import edu.sc.seis.seisFile.mseed.Btime;
import edu.sc.seis.seisFile.mseed.DataHeader;
import edu.sc.seis.seisFile.mseed.DataRecord;
import edu.sc.seis.seisFile.mseed.SeedFormatException;
import edu.sc.seis.seisFile.mseed.SeedRecord;
import edu.sc.seis.seisFile.mseed.Utility;

public class BudLightWriter implements Runnable {

    public BudLightWriter(File dataDir) {
        this(dataDir, new DataRecordQueue());
    }

    public BudLightWriter(File dataDir, DataRecordQueue inQueue) {
        this.inQueue = inQueue;
        this.dataDir = dataDir;
        dataDir.mkdirs();
    }

    public void run() {
        while (true) {
            try {
                if (inQueue.getSize() != 0) {
                    List<DataRecord> drList = inQueue.getWork();
                    outputToFile(drList);
                } else {
                    // queue is empty, so can reprocess some ooo files
                    Iterator<String> it = oooFilenames.iterator();
                    if (it.hasNext()) {
                        String filename = oooFilenames.iterator().next();
                        fixOutOfOrder(filename, EMPTY_LIST);
                    } else {
                        // queue was size 0 and no ooo to process, sleep a little to give queue time to get bigger
                        // as we are more efficient with a queue with some data
                        try {
                            Thread.sleep(100);
                        } catch (InterruptedException e) {}
                    }
                }
            } catch(Exception e) {
                logger.error("exception", e);
            }
        }
    }

    public DataRecordQueue getQueue() {
        return inQueue;
    }

    public File getDataDir() {
        return dataDir;
    }

    public static String calcFilename(DataHeader head) {
        Btime start = head.getStartBtime();
        String net = head.getNetworkCode().trim();
        String sta = head.getStationIdentifier().trim();
        String loc = head.getLocationIdentifier().trim();
        String chan = head.getChannelIdentifier().trim();
        if (head.getDataQualityFlags() != BudLightCachingDC.MISSING_PADDED_DATA) {
            return formatFile(net, sta, loc, chan, start.year, start.jday, start.hour);
        } else {
            // missing data record
            return formatFile(net, sta, loc, chan, start.year, start.jday, start.hour)+"nodata";
        }
    }

    protected void outputToFile(List<DataRecord> drList) throws IOException, SeedFormatException {
        DataRecord first = drList.get(0);
        Iterator<DataRecord> it = drList.iterator();
        while (it.hasNext()) {
            count++;
            DataRecord dr = it.next();
            DataHeader head = dr.getHeader();
            String filename = calcFilename(head).intern();
            synchronized(filename) {
                // occassionally the BudLightWriter and the BudLightDataCenter both try to work on the
                // same file at the same time, updating while it is being read, resulting in errors
                // or possibly corrupted data. By sync-ing on the interned String filename, we keep
                // the file only in use by one at a time
            File f = new File(dataDir, filename);
            MicroSecondDate drStart = FissuresConvert.getMicroSecondTime(head.getStartBtime());
            MicroSecondDate drEnd = FissuresConvert.getMicroSecondTime(head.getLastSampleBtime());
            if (!fileBegin.containsKey(filename) && f.exists()) {
                // maybe from a previous run or older
                loadBeginEnd(filename);
            }
            if (fileBegin.containsKey(filename)) {
                // check if we append or have to resort
                if (drStart.after(fileEnd.get(filename))) {
                    // append
                    DataOutputStream dos;
                    if (filename.equals(lastFilename) && lastDOS != null) {
                        dos = lastDOS;
                    } else {
                        if (lastDOS != null) {
                            lastDOS.close();
                        }
                        dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(f, true)));
                    }
                    dr.write(dos);
                    lastDOS = dos;
                    lastFilename = filename;
                } else {
                    logger.warn(count + " "+inQueue.getSize()+" DataRecord not at end of mseed file, reprocess file   " + filename + "  dr="
                                + drStart+" to "+head.getEndTime() + " < fileBegin="+fileBegin.get(filename)+" fileend=" + fileEnd.get(filename));
                    if (inQueue.getSize() > inQueue.maxQueueSize*.80f) {
                        // over 80% of max, send it to the ooo file
                        outOfOrderDataRecord(dr, filename, drStart);
                        break;
                    }
                    if (filename.equals(lastFilename) && lastDOS != null) {
                        lastDOS.close();
                        lastDOS = null;
                        lastFilename = "";
                    }
                    List<DataRecord> drFromFileList = new ArrayList<DataRecord>();
                    drFromFileList.add(dr);
                    dr = null;
                    while(it.hasNext()) {
                        drFromFileList.add(it.next());
                    }
                    fixOutOfOrder(filename, drFromFileList);
                }
            } else {
                // brand new file
                if (lastDOS != null) {
                    lastDOS.close();
                }
                f.getParentFile().mkdirs();
                DataOutputStream dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(f, true)));
                dr.write(dos);
                dos.flush();
                lastDOS = dos;
                lastFilename = filename;
            }
            filenames.add(filename);
            if (!fileBegin.containsKey(filename) || fileBegin.get(filename).after(drStart)) {
                fileBegin.put(filename, drStart);
            }
            if (!fileEnd.containsKey(filename) || fileEnd.get(filename).before(drEnd)) {
                fileEnd.put(filename, drEnd);
            }
            if (filenames.size() > 500000) {
                String oldest = filenames.remove();
                fileBegin.remove(oldest);
                fileEnd.remove(oldest);
            }
            }
        }
        if (lastDOS != null) {
            lastDOS.flush();
        }
        if (count % 100 == 0) {
            Btime start = first.getHeader().getStartBtime();
            logger.debug(count + " " + inQueue.getSize() + "  " + calcFilename(first.getHeader()) + "  "
                    + twoDigit.format(start.min) + ":" + twoDigit.format(start.sec) + "."
                    + fourDigit.format(start.tenthMilli));
        }
    }

    List<DataRecord> readAllFromFile(File f) throws SeedFormatException, IOException {
        List<DataRecord> out = new ArrayList<DataRecord>();
        DataInputStream dis = new DataInputStream(new BufferedInputStream(new FileInputStream(f)));
        SeedRecord sr;
        while (true) {
            try {
                sr = SeedRecord.read(dis);
            } catch (EOFException ee) {
                // done 
                break;
            }
            // should only be DataRecords
            if (sr instanceof DataRecord) {
                out.add( (DataRecord)sr);
            }
        }
        dis.close();
        return out;
    }
    
    void outOfOrderDataRecord(DataRecord dr, String filename, MicroSecondDate drStart) throws IOException {
        // uh oh
        logger.warn(count+" DataRecord not at end of mseed file, skipping   " + filename + "  dr=" + drStart + " < fileend="+fileEnd.get(filename));
        File oooFile = new File(dataDir, filename+".ooo");
        DataOutputStream outOfOrder = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(oooFile, true)));
        dr.write(outOfOrder);
        outOfOrder.close();
        oooFilenames.add(filename);
    }
    
    void loadBeginEnd(String mseed) throws IOException, SeedFormatException {
        DataInputStream dataInStream = new DataInputStream(new BufferedInputStream(new FileInputStream(new File(dataDir,
                                                                                                                mseed)),
                                                                                   1024));
        DataRecord first = (DataRecord)SeedRecord.read(dataInStream, 4096);
        DataRecord last = first;
        while (true) {
            try {
                last = (DataRecord)SeedRecord.read(dataInStream, first.getRecordSize());
            } catch(EOFException e) {
                break;
            }
        }
        MicroSecondDate firstStart = FissuresConvert.getMicroSecondTime(first.getHeader().getStartBtime());
        MicroSecondDate lastEnd = FissuresConvert.getMicroSecondTime(last.getHeader().getLastSampleBtime());
        fileBegin.put(mseed, firstStart);
        fileEnd.put(mseed, lastEnd);
        dataInStream.close();
    }

    void fixOutOfOrder(String filename, List<DataRecord> extras) throws SeedFormatException, IOException {
        File f = new File(dataDir, filename);
        List<DataRecord> drList = new ArrayList<DataRecord>();
        drList.addAll(readAllFromFile(f));
        drList.addAll(extras);
        f.delete();
        fileBegin.remove(filename);
        fileEnd.remove(filename);
        File oooFile = new File(dataDir, filename+".ooo");
        if (oooFile.exists()) {
            drList.addAll(readAllFromFile(oooFile));
            oooFile.delete();
        }
        
        Collections.sort(drList, new DataRecordBeginComparator());
        // read file and re-process via recursion
        Utility.cleanDuplicatesOverlaps(drList);
        logger.debug(drList.size()+" recursion on "+filename);
        outputToFile(drList);
    }

    private int count = 0;

    private DataOutputStream lastDOS = null;

    private String lastFilename = null;

    public static String formatFile(String net, String sta, String loc, String chan, int year, int jday, int hour) {
        net = net.trim();
        sta = sta.trim();
        loc = loc.trim();
        chan = chan.trim();
        return net + "/" + sta + "/" + year + "/" + threeDigit.format(jday) + "/" + net + "." + sta + "." + loc + "."
                + chan + "_" + year + "_" + threeDigit.format(jday) + "_" + twoDigit.format(hour) + ".mseed";
    }

    public static final DecimalFormat twoDigit = new DecimalFormat("00");

    public static final DecimalFormat threeDigit = new DecimalFormat("000");

    public static final DecimalFormat fourDigit = new DecimalFormat("0000");

    static final List<DataRecord> EMPTY_LIST = Collections.emptyList();
    
    File dataDir;

    DataRecordQueue inQueue;

    LinkedList<String> filenames = new LinkedList<String>();
    
    Set<String> oooFilenames = new HashSet<String>();

    HashMap<String, MicroSecondDate> fileBegin = new HashMap<String, MicroSecondDate>();

    HashMap<String, MicroSecondDate> fileEnd = new HashMap<String, MicroSecondDate>();

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(BudLightWriter.class);
}
